@@ -54,15 +54,15 @@ bytes [#bytes]_                          binary         both
    subtype 0. It will be decoded back to bytes.
 """
 
-import calendar
 import datetime
 import itertools
+import os
 import re
 import struct
 import sys
 import uuid
-from codecs import utf_8_decode as _utf_8_decode  # type: ignore[attr-defined]
-from codecs import utf_8_encode as _utf_8_encode  # type: ignore[attr-defined]
+from codecs import utf_8_decode as _utf_8_decode
+from codecs import utf_8_encode as _utf_8_encode
 from collections import abc as _abc
 from typing import (
     IO,
@@ -100,9 +100,18 @@ from bson.code import Code
 from bson.codec_options import (
     DEFAULT_CODEC_OPTIONS,
     CodecOptions,
+    DatetimeConversion,
     _DocumentType,
     _raw_document_class,
 )
+from bson.datetime_ms import (
+    EPOCH_AWARE,
+    EPOCH_NAIVE,
+    DatetimeMS,
+    _datetime_to_millis,
+    _millis_to_datetime,
+    utc,
+)
 from bson.dbref import DBRef
 from bson.decimal128 import Decimal128
 from bson.errors import InvalidBSON, InvalidDocument, InvalidStringData
@@ -113,14 +122,12 @@ from bson.objectid import ObjectId
 from bson.regex import Regex
 from bson.son import RE_TYPE, SON
 from bson.timestamp import Timestamp
-from bson.tz_util import utc
 
 # Import some modules for type-checking only.
 if TYPE_CHECKING:
     from array import array
     from mmap import mmap
 
-
 try:
     from bson import _cbson  # type: ignore[attr-defined]
 
@@ -187,12 +194,10 @@ __all__ = [
     "is_valid",
     "BSON",
     "has_c",
+    "DatetimeConversion",
+    "DatetimeMS",
 ]
 
-EPOCH_AWARE = datetime.datetime.fromtimestamp(0, utc)
-EPOCH_NAIVE = datetime.datetime.utcfromtimestamp(0)
-
-
 BSONNUM = b"\x01"  # Floating point
 BSONSTR = b"\x02"  # UTF-8 string
 BSONOBJ = b"\x03"  # Embedded document
@@ -413,7 +418,7 @@ def _get_boolean(
 
 def _get_date(
     data: Any, view: Any, position: int, dummy0: int, opts: CodecOptions, dummy1: Any
-) -> Tuple[datetime.datetime, int]:
+) -> Tuple[Union[datetime.datetime, DatetimeMS], int]:
     """Decode a BSON datetime to python datetime.datetime."""
     return _millis_to_datetime(_UNPACK_LONG_FROM(data, position)[0], opts), position + 8
 
@@ -514,19 +519,32 @@ _ELEMENT_GETTER: Dict[int, Callable[..., Tuple[Any, int]]] = {
 if _USE_C:
 
     def _element_to_dict(
-        data: Any, view: Any, position: int, obj_end: int, opts: CodecOptions
+        data: Any,
+        view: Any,
+        position: int,
+        obj_end: int,
+        opts: CodecOptions,
+        raw_array: bool = False,
     ) -> Any:
-        return _cbson._element_to_dict(data, position, obj_end, opts)
+        return _cbson._element_to_dict(data, position, obj_end, opts, raw_array)
 
 else:
 
     def _element_to_dict(
-        data: Any, view: Any, position: int, obj_end: int, opts: CodecOptions
+        data: Any,
+        view: Any,
+        position: int,
+        obj_end: int,
+        opts: CodecOptions,
+        raw_array: bool = False,
     ) -> Any:
         """Decode a single key, value pair."""
         element_type = data[position]
         position += 1
         element_name, position = _get_c_string(data, view, position, opts)
+        if raw_array and element_type == ord(BSONARR):
+            _, end = _get_object_size(data, position, len(data))
+            return element_name, view[position : end + 1], end + 1
         try:
             value, position = _ELEMENT_GETTER[element_type](
                 data, view, position, obj_end, opts, element_name
@@ -545,20 +563,30 @@ else:
 _T = TypeVar("_T", bound=MutableMapping[Any, Any])
 
 
-def _raw_to_dict(data: Any, position: int, obj_end: int, opts: CodecOptions, result: _T) -> _T:
+def _raw_to_dict(
+    data: Any, position: int, obj_end: int, opts: CodecOptions, result: _T, raw_array: bool = False
+) -> _T:
     data, view = get_data_and_view(data)
-    return _elements_to_dict(data, view, position, obj_end, opts, result)
+    return _elements_to_dict(data, view, position, obj_end, opts, result, raw_array=raw_array)
 
 
 def _elements_to_dict(
-    data: Any, view: Any, position: int, obj_end: int, opts: CodecOptions, result: Any = None
+    data: Any,
+    view: Any,
+    position: int,
+    obj_end: int,
+    opts: CodecOptions,
+    result: Any = None,
+    raw_array: bool = False,
 ) -> Any:
     """Decode a BSON document into result."""
     if result is None:
         result = opts.document_class()
     end = obj_end - 1
     while position < end:
-        key, value, position = _element_to_dict(data, view, position, obj_end, opts)
+        key, value, position = _element_to_dict(
+            data, view, position, obj_end, opts, raw_array=raw_array
+        )
         result[key] = value
     if position != obj_end:
         raise InvalidBSON("bad object or element length")
@@ -621,7 +649,7 @@ def _make_c_string_check(string: Union[str, bytes]) -> bytes:
     else:
         if "\x00" in string:
             raise InvalidDocument("BSON keys / regex patterns must not contain a NUL character")
-        return cast(bytes, _utf_8_encode(string)[0]) + b"\x00"
+        return _utf_8_encode(string)[0] + b"\x00"
 
 
 def _make_c_string(string: Union[str, bytes]) -> bytes:
@@ -633,7 +661,7 @@ def _make_c_string(string: Union[str, bytes]) -> bytes:
         except UnicodeError:
             raise InvalidStringData("strings in documents must be valid UTF-8: %r" % string)
     else:
-        return cast(bytes, _utf_8_encode(string)[0]) + b"\x00"
+        return _utf_8_encode(string)[0] + b"\x00"
 
 
 def _make_name(string: str) -> bytes:
@@ -641,7 +669,7 @@ def _make_name(string: str) -> bytes:
     # Keys can only be text in python 3.
     if "\x00" in string:
         raise InvalidDocument("BSON keys / regex patterns must not contain a NUL character")
-    return cast(bytes, _utf_8_encode(string)[0]) + b"\x00"
+    return _utf_8_encode(string)[0] + b"\x00"
 
 
 def _encode_float(name: bytes, value: float, dummy0: Any, dummy1: Any) -> bytes:
@@ -724,6 +752,12 @@ def _encode_datetime(name: bytes, value: datetime.datetime, dummy0: Any, dummy1:
     return b"\x09" + name + _PACK_LONG(millis)
 
 
+def _encode_datetime_ms(name: bytes, value: DatetimeMS, dummy0: Any, dummy1: Any) -> bytes:
+    """Encode datetime.datetime."""
+    millis = int(value)
+    return b"\x09" + name + _PACK_LONG(millis)
+
+
 def _encode_none(name: bytes, dummy0: Any, dummy1: Any, dummy2: Any) -> bytes:
     """Encode python None."""
     return b"\x0A" + name
@@ -814,6 +848,7 @@ _ENCODERS = {
     bool: _encode_bool,
     bytes: _encode_bytes,
     datetime.datetime: _encode_datetime,
+    DatetimeMS: _encode_datetime_ms,
     dict: _encode_mapping,
     float: _encode_float,
     int: _encode_int,
@@ -948,27 +983,6 @@ if _USE_C:
     _dict_to_bson = _cbson._dict_to_bson  # noqa: F811
 
 
-def _millis_to_datetime(millis: int, opts: CodecOptions) -> datetime.datetime:
-    """Convert milliseconds since epoch UTC to datetime."""
-    diff = ((millis % 1000) + 1000) % 1000
-    seconds = (millis - diff) // 1000
-    micros = diff * 1000
-    if opts.tz_aware:
-        dt = EPOCH_AWARE + datetime.timedelta(seconds=seconds, microseconds=micros)
-        if opts.tzinfo:
-            dt = dt.astimezone(opts.tzinfo)
-        return dt
-    else:
-        return EPOCH_NAIVE + datetime.timedelta(seconds=seconds, microseconds=micros)
-
-
-def _datetime_to_millis(dtm: datetime.datetime) -> int:
-    """Convert datetime to milliseconds since epoch UTC."""
-    if dtm.utcoffset() is not None:
-        dtm = dtm - dtm.utcoffset()  # type: ignore
-    return int(calendar.timegm(dtm.timetuple()) * 1000 + dtm.microsecond // 1000)
-
-
 _CODEC_OPTIONS_TYPE_ERROR = TypeError("codec_options must be an instance of CodecOptions")
 
 
@@ -1127,14 +1141,44 @@ def _decode_selective(rawdoc: Any, fields: Any, codec_options: Any) -> Mapping[A
     return doc
 
 
+def _array_of_documents_to_buffer(view: memoryview) -> bytes:
+    # Extract the raw bytes of each document.
+    position = 0
+    _, end = _get_object_size(view, position, len(view))
+    position += 4
+    buffers: List[memoryview] = []
+    append = buffers.append
+    while position < end - 1:
+        # Just skip the keys.
+        while view[position] != 0:
+            position += 1
+        position += 1
+        obj_size, _ = _get_object_size(view, position, end)
+        append(view[position : position + obj_size])
+        position += obj_size
+    if position != end:
+        raise InvalidBSON("bad object or element length")
+    return b"".join(buffers)
+
+
+if _USE_C:
+    _array_of_documents_to_buffer = _cbson._array_of_documents_to_buffer  # noqa: F811
+
+
 def _convert_raw_document_lists_to_streams(document: Any) -> None:
+    """Convert raw array of documents to a stream of BSON documents."""
     cursor = document.get("cursor")
-    if cursor:
-        for key in ("firstBatch", "nextBatch"):
-            batch = cursor.get(key)
-            if batch:
-                stream = b"".join(doc.raw for doc in batch)
-                cursor[key] = [stream]
+    if not cursor:
+        return
+    for key in ("firstBatch", "nextBatch"):
+        batch = cursor.get(key)
+        if not batch:
+            continue
+        data = _array_of_documents_to_buffer(batch)
+        if data:
+            cursor[key] = [data]
+        else:
+            cursor[key] = []
 
 
 def _decode_all_selective(data: Any, codec_options: CodecOptions, fields: Any) -> List[Any]:
@@ -1308,7 +1352,7 @@ class BSON(bytes):
         """
         return cls(encode(document, check_keys, codec_options))
 
-    def decode(self, codec_options: CodecOptions = DEFAULT_CODEC_OPTIONS) -> _DocumentType:  # type: ignore[override]
+    def decode(self, codec_options: "CodecOptions[_DocumentType]" = DEFAULT_CODEC_OPTIONS) -> _DocumentType:  # type: ignore[override,assignment]
         """Decode this BSON data.
 
         By default, returns a BSON document represented as a Python
@@ -1345,3 +1389,16 @@ class BSON(bytes):
 def has_c() -> bool:
     """Is the C extension installed?"""
     return _USE_C
+
+
+def _after_fork():
+    """Releases the ObjectID lock child."""
+    if ObjectId._inc_lock.locked():
+        ObjectId._inc_lock.release()
+
+
+if hasattr(os, "register_at_fork"):
+    # This will run in the same thread as the fork was called.
+    # If we fork in a critical region on the same thread, it should break.
+    # This is fine since we would never call fork directly from a critical region.
+    os.register_at_fork(after_in_child=_after_fork)
